(***********************************************************************)
(*                                                                     *)
(*                              Ref-indic                              *)
(*                                                                     *)
(*                        Refinement indicators                        *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2013-2014 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Hend Ben Ameur <hbenameur@yahoo.ca>                                *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: umatrix_structural_mapping.ml,v 1.12 2014-02-26 23:33:54 fclement Exp $ *)

(* Matrix structural mappings. *)

module Nat = Unatural;;
module Dgraph = Udynamic_graph;;
module Ddgraph = Udelayed_dgraph;;

module Idx = Uindex_index;;
module Index = Umatrix_index;;
module Uidx = Uindex_universe;;
module Univ = Umatrix_universe;;
module Set = Umatrix_nonempty_set;;
module Mapping = Umatrix_graph;;

type 'a mapping = 'a Mapping.t
and 'a sub_mapping = 'a Mapping.t
;;

type 'a partition = 'a sub_mapping array;;

type block_number = Ddgraph.arrow_number;;

type ('a, 'b) block = ('a sub_mapping, 'b) Ddgraph.delayed_arrow;;

type ('a, 'b, 'c) t = ('a sub_mapping, 'b, 'c) Ddgraph.t;;

let make f = Ddgraph.make f Nat._1_;;

let same = Ddgraph.same;;

(* Access functions. *)

let get = Ddgraph.get;;
let set = Ddgraph.set;;
let set_next = Ddgraph.set_next;;

let last_block_number msm =
  Nat.pred (Dgraph.next_unset msm.Ddgraph.graph)
;;

(* Converters. *)

let restriction
    ({ Mapping.domain = dom; _} as map) { Mapping.domain = sub_dom; _} =
  let ss = Set.subset_of_full dom sub_dom in
  Mapping.restriction map ss
;;

let partition_of_mapping pp map = Array.map (restriction map) pp;;

let set_zn_pt geom i _idx uidx =
  let zn = Nat.of_int i in
  geom.(Index.to_int uidx) <- zn
;;

let set_zn geom i { Mapping.domain = dom; _; } =
  Set.iter (set_zn_pt geom i) dom
;;

let partition_to_geometry pp =
  let nbz = Uarray.length pp in
  if nbz = Nat._0_ then assert false else
  let { Mapping.domain = dom0; _; } = pp.(0) in
  let nbpt = Univ.cardinal dom0.Set.universe in
  let geom = Uarray.make nbpt Nat._0_ in
  Array.iteri (set_zn geom) pp;
  geom
;;

let partition_to_values pp = Array.map Mapping.get_constant pp;;

let partition_to_geometry_and_values pp =
  (partition_to_geometry pp, partition_to_values pp)
;;

let one_zone_partition sz m0 =
  let dom0 = Set.make sz in
  [| Mapping.constant dom0 m0; |]
;;

let full_domain_of_msm msm =
  let { Dgraph.value = map; _; } = get msm Nat._0_ in
  let u = map.Mapping.domain.Set.universe in
  Set.full u
;;

let msm_choice msm =
  let { Dgraph.value = map; _; } = get msm Nat._0_ in
  match map.Mapping.images with
  | Mapping.Constant y -> y
  | Mapping.Images ims -> ims.(0)
;;

let images_of_msm msm dom =
  let y0 = msm_choice msm in
  let ims = Uarray.make (Set.cardinal dom) y0 in
  let f_item _i { Dgraph.value = item_map; _; } =
    match item_map with
    | { Mapping.domain = item_dom;
        images = Mapping.Constant item_y; } ->
        let f _ item_uidx =
          let idx = Uidx.index dom.Set.uindices item_uidx in
          let i = Idx.to_int idx in
          ims.(i) <- item_y in
        Set.iter f item_dom
    | { Mapping.domain = item_dom;
        images = Mapping.Images item_ims; } ->
        let f item_idx item_uidx =
          let idx = Uidx.index dom.Set.uindices item_uidx in
          let i = Idx.to_int idx in
          let item_i = Idx.to_int item_idx in
          ims.(i) <- item_ims.(item_i) in
        Set.iter f item_dom in
  Dgraph.iter f_item msm.Ddgraph.graph;
  ims
;;

let to_mapping msm =
  let dom = full_domain_of_msm msm in
  let ims = images_of_msm msm dom in
  Mapping.from_array dom ims
;;

let to_array = Ddgraph.to_array;;

let of_partition = Ddgraph.of_values;;
let to_partition { Ddgraph.graph = dg; _; } = Dgraph.to_values dg;;

(* Iterators. *)

let iter = Ddgraph.iter;;
let fold_left = Ddgraph.fold_left;;
let partition_iter f = Array.iteri (fun i -> f (Unatural.of_int i));;
let iter_to_do = Ddgraph.iter_to_do;;
let iter_value = Ddgraph.iter_value;;
let fold_left_value = Ddgraph.fold_left_value;;
